/* Copyright (c) 2005 PrimeBase Technologies GmbH
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2005-01-10	Paul McCullagh
 *
 * H&G2JCtL
 */

#include "xt_config.h"

#include "heap_xt.h"
#include "thread_xt.h"

#ifdef xt_heap_new
#undef xt_heap_new
#endif

#ifdef DEBUG
xtPublic XTHeapPtr xt_mm_heap_new(XTThreadPtr self, size_t size, XTFinalizeFunc finalize, u_int line, c_char *file, xtBool track)
#else
xtPublic XTHeapPtr xt_heap_new(XTThreadPtr self, size_t size, XTFinalizeFunc finalize)
#endif
{
	XTHeapPtr	hp;
	
#ifdef DEBUG
	hp = (XTHeapPtr) xt_mm_calloc(self, size, line, file);
	hp->h_track = track;
	if (track)
		printf("HEAP: +1  1 %s:%d\n", file, (int) line);
#else
	hp = (XTHeapPtr) xt_calloc(self, size);
#endif
	if (!hp)
		return NULL;

	try_(a) {
		xt_init_rwlock(self, &hp->h_lock);
	}
	catch_(a) {
		xt_free(self, hp);
		throw_();
	}
	cont_(a);

	hp->h_ref_count = 1;
	hp->h_finalize = finalize;
	hp->h_onrelease = NULL;
	return hp;
}

xtPublic void xt_check_heap(XTThreadPtr self, XTHeapPtr hp)
{
#ifdef DEBUG
	xt_mm_malloc_size(self, hp);
#endif
}

#ifdef DEBUG
xtPublic void xt_mm_heap_reference(XTThreadPtr self, XTHeapPtr hp, u_int line, c_char *file)
#else
xtPublic void xt_heap_reference(XTThreadPtr self, XTHeapPtr hp)
#endif
{
	xt_w_lock(self, &hp->h_lock);
	hp->h_ref_count++;
#ifdef DEBUG
	if (hp->h_track)
		printf("HEAP: +1 %2d %s:%d\n", (int) hp->h_ref_count, file, (int) line);
#endif
	xt_rw_unlock(self, &hp->h_lock);
}

xtPublic void xt_heap_release(XTThreadPtr self, XTHeapPtr hp)
{	
	if (!hp)
		return;
#ifdef DEBUG
	xt_r_lock(NULL, &hp->h_lock);
	ASSERT(hp->h_ref_count != 0);
	xt_rw_unlock(self, &hp->h_lock);
#endif
	xt_w_lock(self, &hp->h_lock);
	if (hp->h_onrelease)
		(*hp->h_onrelease)(self, hp);
	if (hp->h_ref_count > 0) {
#ifdef DEBUG
	if (hp->h_track)
		printf("HEAP: -1 %2d\n", (int) hp->h_ref_count);
#endif
		hp->h_ref_count--;
		if (hp->h_ref_count == 0) {
			if (hp->h_finalize)
				(*hp->h_finalize)(self, hp);
			xt_rw_unlock(self, &hp->h_lock);
			xt_free(self, hp);
			return;
		}
	}
	xt_rw_unlock(self, &hp->h_lock);
}

xtPublic void xt_heap_set_release_callback(XTThreadPtr self, XTHeapPtr hp, XTFinalizeFunc onrelease)
{
	hp->h_onrelease = onrelease;
}

xtPublic u_int xt_heap_get_ref_count(struct XTThread *self, XTHeapPtr hp)
{
	return hp->h_ref_count;
}

xtPublic void *xt_heap_rlock(XTThreadPtr th, XTHeapPtr hp)
{
	return xt_r_lock(th, &hp->h_lock) ? hp : NULL;
}

xtPublic void *xt_heap_wlock(XTThreadPtr self, XTHeapPtr hp)
{
	return xt_w_lock(self, &hp->h_lock) ? hp : NULL;
}

xtPublic void xt_heap_unlock(XTThreadPtr self, XTHeapPtr hp)
{
	xt_rw_unlock(self, &hp->h_lock);
}

