/* Copyright (c) 2005 PrimeBase Technologies GmbH
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2005-02-07	Paul McCullagh
 *
 * H&G2JCtL
 */

#include "xt_config.h"

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#include "trace_xt.h"
#include "thread_xt.h"

static xtBool			trace_initialized = FALSE;
static xt_mutex_type	trace_mutex;
static size_t			trace_log_size;
static size_t			trace_log_offset;
static char				*trace_log_buffer;
static u_long			trace_stat_count;

#define DEFAULT_TRACE_LOG_SIZE		5000000

xtPublic xtBool xt_init_trace(void)
{
	int err;

	err = xt_p_mutex_init(&trace_mutex, NULL);
	if (err) {
		xt_log_errno(XT_NS_CONTEXT, err);
		trace_initialized = FALSE;
		return FALSE;
	}
	trace_initialized = TRUE;
	trace_log_buffer = (char *) malloc(DEFAULT_TRACE_LOG_SIZE);
	if (!trace_log_buffer) {
		xt_log_errno(XT_NS_CONTEXT, ENOMEM);
		xt_exit_trace();
		return FALSE;
	}
	trace_log_size = DEFAULT_TRACE_LOG_SIZE;
	trace_log_offset = 0;
	return TRUE;
}

xtPublic void xt_exit_trace(void)
{
	if (trace_initialized) {
#ifdef DEBUG
		xt_dump_trace();
#endif
		xt_free_mutex(&trace_mutex);
		trace_initialized = FALSE;
		if (trace_log_buffer)
			free(trace_log_buffer);
		trace_log_buffer = NULL;
		trace_log_size = 0;
		trace_log_offset = 0;
		trace_stat_count = 0;
	}
}

xtPublic void xt_trace_query(XTThreadPtr self, char *query)
{
	int		qlen = strlen(query), tlen;
	char	*ptr, *qptr;

	if (!self)
		self = xt_get_self();

	xt_lock_mutex_ns(&trace_mutex);

	if (trace_log_offset + qlen + 100 >= trace_log_size) {
		size_t	mlen = trace_log_offset / 2;

		if (mlen < qlen + 100)
			mlen = qlen + 100;
		memmove(trace_log_buffer, trace_log_buffer + trace_log_offset - mlen, mlen+1);
		trace_log_offset = mlen;
	}
/*DBG
	trace_stat_count++;
	tlen = (size_t) sprintf(trace_log_buffer + trace_log_offset, "%lu %s: ", trace_stat_count, self->t_name);
DBG*/
	tlen = (size_t) sprintf(trace_log_buffer + trace_log_offset, "// %lu ", (u_long) time(NULL));
	trace_log_offset += tlen;

	ptr = trace_log_buffer + trace_log_offset;
	qlen = 0;
	qptr = query;
	while (*qptr) {
		if (*qptr == '\n' || *qptr == '\r')
			*ptr = ' ';
		else
			*ptr = *qptr;
		if (*qptr == '\n' || *qptr == '\r' || *qptr == ' ') {
			qptr++;
			while (*qptr == '\n' || *qptr == '\r' || *qptr == ' ')
				qptr++;				
		}
		else
			qptr++;
		ptr++;
		qlen++;
	}

	trace_log_offset += qlen;
	*(trace_log_buffer + trace_log_offset) = '\n';
	*(trace_log_buffer + trace_log_offset + 1) = '\0';
	trace_log_offset++;
	
	xt_unlock_mutex_ns(&trace_mutex);
}

xtPublic void xt_tracef_query(XTThreadPtr self, char *fmt, ...)
{
	va_list	ap;
	size_t	len;

	va_start(ap, fmt);
	xt_lock_mutex_ns(&trace_mutex);

	if (trace_log_offset + 2000 > trace_log_size) {
		size_t	len = trace_log_offset / 2;

		memmove(trace_log_buffer, trace_log_buffer + trace_log_offset - len, len+1);
		trace_log_offset = len;
	}

	trace_stat_count++;
	len = (size_t) sprintf(trace_log_buffer + trace_log_offset, "%lu %s: ", trace_stat_count, self->t_name);
	trace_log_offset += len;
	len = (size_t) vsnprintf(trace_log_buffer + trace_log_offset, trace_log_size - trace_log_offset, fmt, ap);
	trace_log_offset += len;

	xt_unlock_mutex_ns(&trace_mutex);
	va_end(ap);
}

xtPublic void xt_trace(char *fmt, ...)
{
	va_list	ap;
	size_t	len;

	va_start(ap, fmt);
	xt_lock_mutex_ns(&trace_mutex);

	if (trace_log_offset + 2000 > trace_log_size) {
		size_t	len = trace_log_offset / 2;

		/*
		printf("%s", trace_log_buffer);
		trace_log_offset = 0;
		*trace_log_buffer = 0;
		*/

		memmove(trace_log_buffer, trace_log_buffer + trace_log_offset - len, len+1);
		trace_log_offset = len;
	}

	len = (size_t) vsnprintf(trace_log_buffer + trace_log_offset, trace_log_size - trace_log_offset, fmt, ap);
	trace_log_offset += len;

	xt_unlock_mutex_ns(&trace_mutex);
	va_end(ap);
}

xtPublic void xt_trace_now(char *fmt, ...)
{
	va_list		ap;
	size_t		len;
	time_t		ticks;
	struct tm	ltime;

	va_start(ap, fmt);
	xt_lock_mutex_ns(&trace_mutex);

	if (trace_log_offset + 2040 > trace_log_size) {
		size_t	len = trace_log_offset / 2;

		/*
		printf("%s", trace_log_buffer);
		trace_log_offset = 0;
		*trace_log_buffer = 0;
		*/

		memmove(trace_log_buffer, trace_log_buffer + trace_log_offset - len, len+1);
		trace_log_offset = len;
	}

	ticks = time(NULL);
	localtime_r(&ticks, &ltime);
	len = strftime(trace_log_buffer + trace_log_offset, 1000, "%Y-%m-%d %H:%M:%S ", &ltime);
	trace_log_offset += len;

	len = (size_t) vsnprintf(trace_log_buffer + trace_log_offset, trace_log_size - trace_log_offset, fmt, ap);
	trace_log_offset += len;

	xt_unlock_mutex_ns(&trace_mutex);
	va_end(ap);
}


xtPublic void xt_dump_trace(void)
{
	FILE *fp;

	if (trace_log_offset) {
		fp = fopen("pbxt.log", "w");

		xt_lock_mutex_ns(&trace_mutex);
		/*
		printf("%s", trace_log_buffer);
		*/
		if (fp) {
			fprintf(fp, "%s", trace_log_buffer);
			fclose(fp);
		}
		trace_log_offset = 0;
		*trace_log_buffer = 0;
		xt_unlock_mutex_ns(&trace_mutex);
	}
}

xtPublic xtWord8 xt_trace_clock(void)
{
	static xtWord8	trace_start_clock = 0;
	struct timeval	tv;
	xtWord8			now;

	gettimeofday(&tv, NULL);
	now = (xtWord8) tv.tv_sec * (xtWord8) 1000000 + tv.tv_usec;
	if (trace_start_clock)
		return now - trace_start_clock;
	trace_start_clock = now;
	return 0;
}

xtPublic char *xt_trace_clock_str(char *ptr)
{
	static char	buffer[50];
	xtWord8		now = xt_trace_clock();

	if (!ptr)
		ptr = buffer;

	sprintf(ptr, "%d.%06d", (int) (now / (xtWord8) 1000000), (int) (now % (xtWord8) 1000000));
	return ptr;
}

xtPublic char *xt_trace_clock_diff(char *ptr)
{
	static xtWord8	trace_last_clock = 0;
	static char		buffer[50];
	xtWord8			now = xt_trace_clock();

	if (!ptr)
		ptr = buffer;

	sprintf(ptr, "%d.%06d (%d)", (int) (now / (xtWord8) 1000000), (int) (now % (xtWord8) 1000000), (int) (now - trace_last_clock));
	trace_last_clock = now;
	return ptr;
}

